""" 
Utility functions for icat module. 

@author: Dakota Y. Hawkins
@contact: dyh0110@bu.edu 
"""
import collections
import inspect
import sys
import os
import re
import warnings
import logging
import time
from pkg_resources import parse_version

import psutil
import numpy as np
import pandas as pd
import scanpy as sc
from scipy import spatial
from sklearn import model_selection

import igraph as ig

from ncfs import distances


def ftime(seconds):
    return time.strftime("%Hh:%Mm:%Ss", time.gmtime(seconds))


def set_log():
    # logging.basicConfig(stream=sys.stdout, level=logging.DEBUG)
    logging.basicConfig(filename="icat.log", level=logging.DEBUG, filemode="w")
    # root = logging.getLogger()
    # root.setLevel(logging.DEBUG)
    # handler = logging.StreamHandler(sys.stdout)
    # handler.setLevel(logging.DEBUG)
    # formatter = logging.Formatter('%(asctime)s\t%(name)s\t%(levelname)s\t%(message)s')
    # handler.setFormatter(formatter)
    # root.addHandler(handler)
    logging.info("System Usage upon startup")
    log_system_usage()


def log_system_usage(msg=None):
    pid = os.getpid()
    py = psutil.Process(pid)
    memory_use = py.memory_info().rss / 2 ** 30
    if msg is not None:
        logging.info(msg)
    logging.info("Memory usage: {:0.03} GB".format(memory_use))


def close_log():  # this doesn't work
    logging.getLogger().close()


def check_kws(reference_dict, new_dict, name):
    if not isinstance(new_dict, dict):
        raise ValueError(
            "Expected dictionary of keyword arguments for "
            "`{}`. Received {}.".format(name, type(new_dict))
        )
    for key, item in new_dict.items():
        if key not in reference_dict.keys():
            raise ValueError(
                "Unsupported keyword argument `{}` for "
                "{} keywords.".format(key, name)
            )
        reference_dict[key] = item
    return reference_dict


def get_default_kwargs(func, ignore_params=[]):
    params = inspect.signature(func).parameters
    kwargs = {x: params[x].default for x in params if x not in ignore_params}
    return kwargs


def check_string_ids(adata):
    if isinstance(adata.obs.index, pd.RangeIndex):
        warnings.warn(
            "WARNING: Numeric index used for cell ids. " "Converting to strings."
        )
        adata.obs.index = adata.obs.index.map(str)
    if isinstance(adata.var.index, pd.RangeIndex):
        warnings.warn(
            "WARNING: Numeric index used for gene ids. " "Converting to strings."
        )
        adata.var.index = adata.var.index.map(str)


def check_matching_genes(ref, new):
    """Ensure two AnnData objects have shared genes."""
    return all(ref.var.index == new.var.index)


def subset_cells(adata, vector, value, copy=True):
    """Subset cells by finding indices in `vector` where `value` occurs."""
    idxs = np.where(vector == value)[0]
    if copy:
        return adata[idxs, :].copy()
    return adata[idxs, :]


def check_np_castable(obj, name):
    """Check whether an object is castable to a numpy.ndarray."""
    if not isinstance(obj, np.ndarray):
        try:
            obj = np.array(obj)
        except:
            raise ValueError(
                "Expected numpy.ndarray castable object for "
                "`{}`. Got {}.".format(name, type(obj))
            )
    return obj


# --------------------------- Cell subselection Method -------------------------
def get_neighbors(adata, measure="connectivities"):
    # grab connectivities of cells
    if measure not in ["connectivities", "distances"]:
        raise ValueError(
            "`measure` should be either 'connectivities or "
            "or 'distances. Received {}.".format(measure)
        )
    if parse_version(sc.__version__) < parse_version("1.5.0"):
        return adata.uns["neighbors"][measure]
    else:
        return adata.obsp[measure]


def distance_matrix(X, metric, kws={}):
    D = np.zeros((X.shape[0], X.shape[0]))
    if isinstance(metric, str):
        try:
            dist_func = distances.supported_distances[metric]
            distances.pdist(X, np.ones(X.shape[1]), D, dist_func)
        except KeyError:
            D = spatial.distance.pdist(X, metric=metric, **kws)
    else:
        distances.pdist(X, np.ones(X.shape[1]), D, metric)
    return D


def assign_selected(adata, indices, subset_indices=None):
    """
    Specify which cells were selected to learn NCFS weights.

    Parameters
    ----------
    adata : sc.AnnData
        Single-cell dataset cells were selected from
    indices : numpy.ndarray
        Array specifying observation indices of selected cells.
    subset_indices : numpy.ndarray, optional
        Used if `adata` is a subset view of a larger dataset. `subset_indices`
        is therefore an array of indices pointing to original indices of the
        view. By default None, and `adata` is assumed to be the complete dataset.

    Returns
    -------
    None
        Creates `selected` column in observation annotations.
    """
    selected = adata.obs.index.values[indices]
    if subset_indices is not None:
        selected = adata.obs.index.values[subset_indices[indices]]
    if "selected" not in adata.obs.columns:
        adata.obs["selected"] = False
    adata.obs.loc[selected, "selected"] = True


def get_data_representation(adata, use_rep):
    if use_rep == "X":
        X = adata.X
    else:
        if use_rep not in adata.obsm.keys():
            raise KeyError(f"Data representation {use_rep} not found.")
        X = adata.obsm[use_rep]
    return X


def submodular_select(
    adata,
    y,
    by_cluster,
    stratified,
    train_size,
    metric,
    selector,
    metric_kwargs={},
    use_rep="X",
    verbose=True,
):
    if verbose:
        print("Selecting training cells using submodular optimization...")
    X = get_data_representation(adata, use_rep)
    if by_cluster:
        if verbose:
            print("Selecting cells for each cluster individually...")
        labels, counts = np.unique(y, return_counts=True)
        train_X = []
        train_y = []
        for label, count in zip(labels, counts):
            subset_idxs = np.where(y == label)[0]
            subset = X[subset_idxs, :]
            D = distance_matrix(subset, metric, metric_kwargs)
            n_samples = int(X.shape[0] / len(labels) * train_size)
            # will have to check to see if n_samples > count
            if stratified:
                n_samples = int(count * train_size)
            if n_samples < subset.shape[0]:
                model = selector(n_samples, metric="precomputed")
                X = model.fit_transform(D)
                train_X.append(X)
                train_y.append([label] * X.shape[0])
            else:
                msg = (
                    "Number of samples to select exceeds number of "
                    "cells in provided cluster. Selecting all "
                    "samples in cluster {}".format(label)
                )
                warnings.warn(msg)
                train_X.append(subset)
                train_y.append([label] * subset.shape[0])
            if verbose:
                print(
                    f"cluster {label} size: {count}, train_size: {train_X[-1].shape[0]}"
                )
            assign_selected(adata, model.ranking, subset_idxs)
        train_X = np.vstack(train_X)
        train_y = np.hstack(train_y)
    else:
        model = selector(int(X.shape[0] * train_size), metric="precomputed")
        D = distance_matrix(X, metric, metric_kwargs)
        model.fit(D)
        train_X = X[model.ranking, :]
        train_y = y[model.ranking]
        assign_selected(adata, model.ranking)
        if verbose:
            print("Selected {} cells".format(train_X.shape[0]))
            labels, counts = np.unique(train_y, return_counts=True)
            for label, count in zip(labels, counts):
                print(f"cluster {label} size: {count}")
    return (train_X, train_y)


def random_select(adata, y, train_size, stratified, use_rep="X", verbose=True):
    X = get_data_representation(adata, use_rep)
    if verbose:
        print("Randomly selecting training cells.")
    if stratified:
        split = model_selection.StratifiedShuffleSplit(
            n_splits=1, train_size=train_size
        )
    else:
        split = model_selection.ShuffleSplit(n_splits=1, train_size=train_size)
    train_idxs, __ = next(split.split(X, y))
    train_X = X[train_idxs, :]
    train_y = y[train_idxs]
    assign_selected(adata, train_idxs)
    return (train_X, train_y)


def centroid_collapse(adata, y, use_rep="X", verbose=True):
    if verbose:
        print("Collapsing clusters to their median centroid")
    if use_rep != "X":
        warnings.warn(
            "`centroid` was passed to `select_cells()`, but "
            "`use_rep` was not set to X. As ncfs finds "
            "informative features, data matrix `X` was used."
        )
    X = adata.X
    train_X = []
    train_y = []
    for label in np.unique(y):
        subset_idxs = np.where(y == label)[0]
        subset = X[subset_idxs, :]
        train_X.append(np.median(subset, axis=0).reshape(1, -1))
        train_y.append([label])
    return (np.vstack(train_X), np.hstack(train_y))


def centroid_neighbors(
    adata,
    y,
    train_size,
    metric,
    stratified,
    metric_kwargs={},
    use_rep="X",
    verbose=True,
):
    if verbose:
        print("Selecting nearest neighbors to cluster centroids.")
    train_X = []
    train_y = []
    X = get_data_representation(adata, use_rep)
    labels, counts = np.unique(y, return_counts=True)
    for label, count in zip(labels, counts):
        subset_idxs = np.where(y == label)[0]
        subset = X[subset_idxs, :]
        centroid = np.median(subset, axis=0).reshape(1, -1)
        n_samples = int(X.shape[0] / len(labels) * train_size)
        if stratified:
            n_samples = int(count * train_size)
        if n_samples > subset.shape[0]:
            train_X.append([subset])
            train_y.append([label] * subset.shape[0])
        if metric == "manhattan":
            metric = "cityblock"
        cdistances = spatial.distance.cdist(
            subset, centroid, metric=metric, **metric_kwargs
        ).flatten()
        neighbors = np.argsort(cdistances)[:n_samples]
        assign_selected(adata, neighbors, subset_idxs)
        train_X.append(subset[neighbors, :])
        train_y.append([label] * n_samples)
        if verbose:
            print(f"Selected {n_samples} cells for cluster {label}")
    return (np.vstack(train_X), np.hstack(train_y))


def __evaluate_key(key, sep):
    if not isinstance(key, str):
        raise ValueError("Keys must be strings for easy concatentation.")
    if sep in key:
        raise ValueError("Cannot have `{}` in dictionary keys.".format(sep))


def flatten_dict(d, parent_key="", sep="."):
    """
    Flatten a dictionary containing nested dictionaries.

    Parameters
    ----------
    d : dict
        Dictionary to flatten
    parent_key : str, optional
        Key in parent dictionary pointing to `d`. The default is '', which
        assumes `d` is the highest level nested dictionary.
    sep : str, optional
        String value to separate child and parent keys. The default is '.',
        which will place a '.' between each key. All parent and child keys
        will be assessed to ensure they do not contain a `sep` character;
        therefore, `sep` should be set to a delimiter not present in current
        keys.

    Returns
    -------
    dict
        Flattened dictionary with parent and child keys separted by `sep`.

    References
    ----------

    Taken shamelessly from here:
        https://stackoverflow.com/questions/6027558/flatten-nested-python-dictionaries-compressing-keys
    """

    items = []
    for k, v in d.items():
        __evaluate_key(k, sep)
        new_key = parent_key + sep + k if parent_key else k
        if isinstance(v, collections.MutableMapping):
            items.extend(flatten_dict(v, new_key, sep=sep).items())
        else:
            items.append((new_key, v))
    return dict(items)


def igraph_from_adjacency(adjacency, directed=None):
    """
    Get igraph graph from adjacency matrix.

    Parameters
    ----------

    adjacency : numpy.ndarray, scipy.sparse.csr
        Adjacency matrix where non-zero entries represent connections between
        samples.
    directed: boolean, optional

    Returns
    -------
    igraph.graph
        Nearest neighbor graph generated from adjacency matrix.

    References
    ----------
    Taken from: https://github.com/theislab/scanpy/blob/28498953092dc7cbecd0bd67380b1b060367d639/scanpy/_utils.py#L170
    """
    import igraph as ig

    sources, targets = adjacency.nonzero()
    weights = adjacency[sources, targets]
    if isinstance(weights, np.matrix):
        weights = weights.A1
    g = ig.Graph(directed=directed)
    g.add_vertices(adjacency.shape[0])  # this adds adjacency.shape[0] vertices
    g.add_edges(list(zip(sources, targets)))
    try:
        g.es["weight"] = weights
    except:
        pass
    if g.vcount() != adjacency.shape[0]:
        warnings.warn(
            f"The constructed graph has only {g.vcount()} nodes. "
            "Your adjacency matrix contained redundant nodes."
        )
    return g


def is_none(x):
    """Check whether a value is a null value."""
    if isinstance(x, float):
        return np.isnan(x)
    if isinstance(x, str):
        return x.lower() in ["none", "nan", "na"]
    return x is None


def format_labels(clusters):
    """
    Format cluster labels for sslouvain.

    Parameters
    ----------
    clusters : iterable
        List of cluster labels where None or np.nan represent previously
        unlabeled samples.

    Returns
    -------
    (list, list):
        labels : list
            List of labels where previously unlabeled samples are given
            their own labels
        mutables : list
            List of samples indicating which samples were previously labelled.
    """
    if isinstance(clusters, pd.Series):
        clusters = clusters.values
    elif isinstance(clusters, list):
        clusters = np.array(clusters)
    elif not isinstance(clusters, np.ndarray):
        warnings.warn(f"Unsupport type {type(clusters)} for `clusters`")
    mutables = [True] * len(clusters)
    labels = [None] * len(clusters)
    named_clusters = set([x for x in clusters if not is_none(x)])
    label_to_int = {x: i for i, x in enumerate(sorted(named_clusters))}
    start_label = len(label_to_int)
    for i, x in enumerate(clusters):
        if is_none(x):
            labels[i] = start_label
            start_label += 1
        else:
            labels[i] = label_to_int[x]
            mutables[i] = False
    return (labels, mutables)



"""
Module to simulate single-cell RNA sequencing data

@author: Dakota Y. Hawkins
@contact: dyh0110@bu.edu 
"""
import re

import numpy as np
import pandas as pd
import scanpy as sc
from scipy import stats

try:
    from . import utils
except ImportError:
    import utils


class SingleCellDataset:
    """
    Class to simulate a single-cell RNA sequencing dataset.

    Attributes
    ----------
    samples : int
        Number of cells to simulate.
    genes : int
        Number of genes to simulate.
    populations : int
        Number of populations to simulate in the dataset.
    pop_sizes : list-like
        Number of cells in each population.
    p_marker : float
        Probability of selecting a gene as a marker gene for a given population.
    dispersion : int, list-like
        Dispersion parameter used to simulate counts in a negative binomial
        model. Either a single, integer value shared across all genes, or an
        array of gene-specific values.
    scalar : float
        Scalar value multiplied to a beta-distributed random variable to
        estimate average gene expression for a simualted gene.
    Methods
    -------

    get_params():
        Get model parameters used to simulate dataset.
    simulate():
        Simulate a single-cell RNA sequencing dataset with the set parameters.
    """

    def __init__(
        self,
        samples=200,
        genes=1000,
        populations=2,
        pop_sizes=None,
        p_marker=None,
        dispersion=1,
        scalar=100,
        percentile=50,
        method="conditional",
        dropout=0.66,
    ):
        """
        Parameters
        ----------
        samples : int, optional
            Number of cells to simulate. Default is 200.
        genes : int, optional
            Number of genes to simulate. Default is 1000.
        populations : int, optional
            Number of populations to simulate in the dataset. Default is 2.
        pop_sizes : list-like, optional
            Number of cells in each population. Default is None, and the number
            of cells will be split uniformly between populations.
        p_marker : float, optional
            Probability of selecting a gene as a marker gene for a given
            population. Default is None, and will be set so that each
            population will have an average of 10 marker genes.
        dispersion : int, list-like, optional
            Dispersion parameter used to simulate counts in a negative binomial
            model. Either a single, integer value shared across all genes, or an
            array of gene-specific values. Default is 1, and will be shared
            across all genes.
        scalar : float, optional
            Scalar value multiplied to a beta-distributed random variable to
            estimate average gene expression for a simualted gene. Default is
            100.
        percentile : float, optional
            Float value between 0 and 100 denoting which percentile to use when
            calculating dropout probabilities. Default is 50, and the median
            will be calculated.
        """
        self.samples = samples
        self.genes = genes
        self.populations = populations
        self.pop_sizes = pop_sizes
        self.p_marker = p_marker
        self.dispersion = dispersion
        self.scalar = scalar
        self.percentile = percentile
        self.method = method
        self.dropout = dropout

    @property
    def samples(self):
        """Get samples attribute."""
        return self._samples

    @samples.setter
    def samples(self, value):
        """Set samples attribute."""
        if value < 0 and int(value) != value:
            raise ValueError("Expected positive integer for sample size.")
        self._samples = value

    @property
    def genes(self):
        """Get genes attribute."""
        return self._genes

    @genes.setter
    def genes(self, value):
        """Set genes attribute."""
        if value < 0 and int(value) != value:
            raise ValueError("Expected positive integer for number of genes.")
        self._genes = value

    @property
    def populations(self):
        """Get populations attribute."""
        return self._populations

    @populations.setter
    def populations(self, value):
        """Set populations attribute."""
        if value < 0 and int(value) != value:
            raise ValueError("Expected positive integer for number of " "populations.")
        self._populations = value

    @property
    def pop_sizes(self):
        """Get pop_sizes attribute."""
        return self._pop_sizes

    @pop_sizes.setter
    def pop_sizes(self, value):
        """Set pop_sizes attribute."""
        if value is None:
            value = np.array(
                [self.samples // self.populations for each in range(self.populations)]
            )
            remainder = self.samples % self.populations
            if remainder != 0:
                value[:remainder] += 1
        try:
            value = np.array(value, dtype=int)
        except TypeError:
            raise ValueError(
                "Expected numpy.array castable when setting " "population sizes."
            )
        except ValueError:
            raise ValueError(
                "Expected numpy.array castable when setting " "population sizes."
            )
        if len(value) != self.populations:
            raise ValueError(
                "Number of populations does not match number of "
                "populations passed. "
                "Populations set to {}".format(self.populations),
                ", but received " "{} population sizes.".format(self.pop_sizes),
            )
        if value.sum() != self.samples:
            raise ValueError(
                "Expected population sizes to total to number of "
                "samples. Got {}".format(value.sum())
            )
        self._pop_sizes = value

    @property
    def p_marker(self):
        """Get p_marker attribute."""
        return self._p_marker

    @p_marker.setter
    def p_marker(self, value):
        """Set p_marker attribute."""
        if value is None:
            value = 10 / self.genes
        elif value < 0 or value > 1:
            raise ValueError(
                "Expected value for p_marker between 0 and 1. " "Got {}.".format(value)
            )
        self._p_marker = value

    @property
    def dispersion(self):
        """Get dispersion parameter."""
        return self._dispersion

    @dispersion.setter
    def dispersion(self, value):
        """Set dispersion parameter."""
        if isinstance(value, (list, np.ndarray)):
            if not len(value) == self.genes:
                raise ValueError(
                    "Number of dispersions passed does not match "
                    "the number of genes passed. Dispersion should"
                    " either be a list-like of length `genes` or "
                    "a single integer value."
                )
            value = np.array(value, dtype=int)
            if not np.all(value > 0):
                raise ValueError(
                    "Expected non-negative positive integer for "
                    "all dispersion parameters."
                )
        elif isinstance(value, (int, np.integer)):
            if value < 1:
                raise ValueError(
                    "Dispersion must be a non-negative integer. "
                    "Received: {}".format(value)
                )
            # cast to array of length `genes`
            value = np.array([value] * self.genes)
        else:
            raise ValueError(
                "Expected integer value for dispersion parameter "
                "Received: {} - {}".format(value, type(value))
            )

        self._dispersion = value

    @property
    def scalar(self):
        """Get scalar parameter."""
        return self._scalar

    @scalar.setter
    def scalar(self, value):
        """Set scalar parameter."""
        if not isinstance(value, (float, int, np.float, np.integer)):
            raise ValueError(
                "Expected numerical value for `scalar` parameter."
                "Received: {}".format(type(value))
            )
        if value < 1:
            raise ValueError(
                "Expected `scalar` value > 1: values less than one"
                " will result in average gene expressions < 1."
            )
        self._scalar = value

    @property
    def percentile(self):
        """Get percentile parameter."""
        return self._percentile

    @percentile.setter
    def percentile(self, value):
        try:
            float(value)
        except:
            raise ValueError("Expected numerical value for `percentile`.")
        if not 0 <= value <= 100:
            raise ValueError("`percentile` must be a float between 0 and 100.")
        self._percentile = value

    @property
    def method(self):
        """
        Which method to use to calculate gene dropout. Options are
        "conditional" and "uniform".
        """
        return self._method

    @method.setter
    def method(self, value):
        if value not in ["conditional", "uniform"]:
            raise ValueError("Unsupported method {}.".format(value))
        self._method = value

    @property
    def dropout(self):
        """
        Average percent of genes to dropout during count simulation. Only used
        if `method == uniform`.
        """
        return self._dropout

    @dropout.setter
    def dropout(self, value):
        if value < 0 or value > 1:
            raise ValueError(
                "Expected values between zero and 1. " "Received {}.".format(value)
            )
        self._dropout = value

    def __repr__(self):
        """Return string representation of SingleCellDataset object."""
        header = "Simulated Single-Cell Dataset\n"
        out = ["{}: {}".format(k, v) for k, v in self.get_params().items()]
        return header + "\n".join(out)

    def get_params(self):
        """Get parameter set used to simulate dataset."""
        out_dict = {
            "samples": self.samples,
            "genes": self.genes,
            "populations": self.populations,
            "pop_sizes": self.pop_sizes,
            "p_marker": self.p_marker,
            "dispersion": self.dispersion,
            "scalar": self.scalar,
            "percentile": self.percentile,
            "method": self.method,
            "dropout": self.dropout,
        }
        return out_dict

    def simulate(self):
        r"""
        Simulate read counts across genes and populations.           
        
        Parameters
        ----------
        None
        
        Returns
        -------
        sc.AnnData
            Annotated dataframe of simulated data.

        Notes
        -----
        Counts are simulated from a negative binomial distribution with dropout.

        ..math::
             c_{ij} ~ NV(r_ij, 1 - \dfrac{\mu_{ij}}{\mu + 1} | Bernoulli(p_ij))

        Where :math:`c_ij` is the read counts for gene :math:`j` in cell
        :math:`i`, :math:`r_ij` is the dispersion parameter for gene `j` when
        expressed in cell `i`, :math:`\mu_{ij}` is the average expression value
        for gene :math:`j` in cell :math:`i`, and :math:`p_ij` is the dropout
        probability of gene :math:`j` in cell :math:`i`. As :math:`c_{ij}` is
        conditioned on :math:`p_{ij}`, if :math:`\not p_{ij}, c_{ij} = 0`. 

        The baseline average expression value for gene `j`, :math:`\mu_{j}`, is
        modelled using a scaled Beta distribution with parameters :math:`a = 2`
        and :math:`b = 5`. Each Beta random variable is scaled by a constant
        :math:`c = 100` to calculate the final average.
        
        ..math::
            \mu_j \sim Beta(a, b) \cdot c

        If gene :math:`j` was selected as a marker gene for population
        :math:`K`, this average is scaled again to represent up or down
        regulation in population :math:`K` compared to baseline. The marker gene
        scalar is drawn from a Gamma-distributed random variable, where:

        ..math::
            \gamma \sim Gamma(\alpha, \beta) \\
            \mu_{j}^K = \gamma \mu_j \\
            
        and :math:`\alpha = 2`, `\beta = 2`. Thus, letting :math:`M` represent
        the set of all marker genes.
            
        ..math::
            \mu_ij = 
            \begin{cases}
                \mu_{j}^K & \forall i \in K \and j \in M`, \\
                \mu_j & otherwise
            \end{cases}

        The probability for a dropout event for gene :math:`j` in cell `i` is
        modelled by passing an affine transformation through a sigmoid function:

        ..math::
            p_ij = sigmoid(x) \\
            sigmoid(x) = \dfrac{1}{1 + e^{-x}} \\
            x_ij = \beta_0 + \dfrac{mu_ij}{median(\vec \mu)} \\
            \vec \mu = \{\mu_1, \mu_1, \ldots , \mu_{p - 1}, \mu_p}
        
        Here, :math:`beta_0 = -1.5` and :math:`median(\vec \mu)` is the median
        average baseline expression value across all genes. That is, shifts
        for marker genes are not taken into consideration.
        """
        # data frame to track cell annotations
        obs = pd.DataFrame(index=range(self.samples), columns=["Population"])
        # data frame to track gene annotations
        var = pd.DataFrame(
            index=range(self.genes),
            columns=["Pop.{}.Marker".format(i + 1) for i in range(self.populations)],
        )
        var.fillna(False, inplace=True)
        var["Base.Dispersion"] = self.dispersion
        # get baseline expression averages
        mus_ = average_exp(scale_factor=self.scalar, n=self.genes)
        var["Base.Mu"] = mus_
        # create gene x pop matrix to hold pop-specific expression averages
        mus = np.ones((mus_.size, self.populations)) * mus_.reshape(-1, 1)
        # get number of marker genes for each population
        n_markers = stats.binom(self.genes, self.p_marker).rvs(self.populations)
        # must have at least 1 marker gene
        n_markers[n_markers == 0] = 1
        # distribution to sample expression shifts from
        gamma = stats.gamma(a=3, scale=3)
        possible_markers = var.index.values
        for i, n in enumerate(n_markers):
            # pick marker genes and shift their expression in population i
            markers = np.random.choice(possible_markers, n)
            mus[markers, i] = mus[markers, i] * gamma.rvs(n)
            # log marker genes in var data frame
            var.loc[markers, "Pop.{}.Marker".format(i + 1)] = True
            # remove selected markers from pool, so that populations don't
            # share marker genes
            possible_markers = np.array(list(set(possible_markers).difference(markers)))

        X, dropout, labels = simulate_counts(
            self.samples,
            mus,
            self.dispersion,
            self.populations,
            self.pop_sizes,
            percentile=self.percentile,
            method=self.method,
            dropout=self.dropout,
        )
        # log population averages and dropout probabilities
        for i in range(self.populations):
            var["Pop.{}.Mu".format(i + 1)] = mus[:, i]
            var["Pop.{}.Dropout".format(i + 1)] = dropout[:, i]

        obs["Population"] = labels
        obs["Population"] = obs["Population"].astype(str)
        obs["Treatment"] = "Control"
        # rename obs and var rows for clarity
        obs.rename(index={i: "cell-{}".format(i + 1) for i in obs.index}, inplace=True)
        var.rename(index={i: "gene-{}".format(i + 1) for i in var.index}, inplace=True)
        return sc.AnnData(X=X, obs=obs, var=var)


class Experiment(object):
    """Class to simulate scRNA experiments with perturbations."""

    def __init__(self, control_kwargs=None, perturb_kwargs=None):
        """
        Simulate a scRNAseq experiment with perturbations.

        Parameters
        ----------
        control_kwargs : dict, optional
            Dictionary of keyword arguments that specify simulation parameters.
            See `SingleCellDataSet` for more infomration. By default None, and
            default parameters for SingleCellDataSet will be used.
        perturb_kwargs : dict, optional
            Dictionary of keyword argument that specify perturbation paramters.
            By default None, and default parameters will be used. See
            `perturb()` for more information.
        """

        self.control_kwargs = control_kwargs
        self.perturb_kwargs = perturb_kwargs

    @property
    def control_kwargs(self):
        """Get control_kwargs attribute."""
        return self._control_kwargs

    @control_kwargs.setter
    def control_kwargs(self, value):
        """Set control_kwargs attribute."""
        default_kws = utils.get_default_kwargs(SingleCellDataset, ["self"])
        if value is not None:
            value = utils.check_kws(default_kws, value, "control_kwargs")
        else:
            value = default_kws
        self._control_kwargs = value

    @property
    def perturb_kwargs(self):
        """Get perturb_kwargs attribute."""
        return self._perturb_kwargs

    @perturb_kwargs.setter
    def perturb_kwargs(self, value):
        """Set perturb_kwargs attribute."""
        default_kws = utils.get_default_kwargs(perturb, ["adata"])
        if value is not None:
            value = utils.check_kws(default_kws, value, "perturb_kwargs")
        else:
            value = default_kws
        if "percentile" not in value and "percentile" in self.control_kwargs:
            value["percentile"] = self.control_kwargs["percentile"]
        self._perturb_kwargs = value

    def simulate_controls(self):
        adata = SingleCellDataset(**self.control_kwargs).simulate()
        return adata

    def run(self, simulations=1, replications=1, controls=None):
        """
        Simulate control and perturbed datasets under experimental conditions.

        Parameters
        ----------
        simulations : int, optional
            Number of control datasets to simulate. Default is 1, and a single
            reference control dataset will be simulated.
        replications : int, optional
            Number of perturbations to simulate for each control dataset.
            Default is 1, and a single perturbation will be simulated for each
            reference control dataset.
        controls : sc.AnnData, optional
            A dataset of simulated control cells to perturb. Default is None,
            and a control dataset will be simulated according to parameters
            defined by `control_kwargs`.

        Returns
        -------
        list
            Two-dimensional list that is indexed first by simulation and second
            by replicate.
        """
        out = []
        for __ in range(simulations):
            sim_out = []
            if controls is None:
                controls = self.simulate_controls()
            if not isinstance(controls, sc.AnnData):
                raise ValueError(
                    "Unexpected type for `controls`: {}".format(type(controls))
                )
            for __ in range(replications):
                treated = perturb(controls, **self.perturb_kwargs)
                combined = controls.concatenate(treated)
                assert set(controls.var.columns).issubset(treated.var.columns)
                combined.var = treated.var
                sim_out.append(combined)
            out.append(sim_out)
        return out


def dispersions(size, a=1, b=5):
    """
    Randomly choose dispersion parameter 'r' for simulating cell counts.

    Parameters
    ----------
    size : int, tuple
        Size of array to create.
    a : int, optional
        Minimum dispersion value, by default 1.
    b : int, optional
        Maximum dispersion value, by default 5.

    Returns
    -------
    numpy.ndarray
        array of dispersion paramters.
    """
    return np.random.choice(range(a, b), size=size)


def population_markers(adata):
    # pop_regex = re.compile('(?<=\.)(.*)(?=\.)')
    # marker_cols = [x for x in adata.var.columns if 'Marker' in x]
    # marker_cols = [f"Pop.{x}.Marker" for x in ]
    markers = {
        x: adata.var[adata.var[f"Pop.{x}.Marker"]].index.values
        for x in adata.obs.Population.unique()
    }
    return markers


def perturb(
    adata,
    samples=200,
    pop_targets=None,
    gene_targets=None,
    percent_perturb=0.2,
    pop_sizes=None,
    percentile=50,
    new_pop_cells=[],
    new_pop_pmarker=None,
    new_pop_ids=None,
    perturbation_key=None,
):
    r"""
    Perturb a simulated single-cell dataset.

    Perturbs a simulated single-cell dataset by applying a Gamma-distributed
    scalar shift to the average expression value to a set number of genes.

    That is, for a specific simulated gene following a negative binomial 
    distribution :math:`gene_i \sim NB(r_i, \mu_i)`, the perturbation is
    modelled using the following equations: 

    .. math::
        \gamma_i \sim Gamma(2, 2) \\
        gene_{i} \sim NB(r_i, \gamma_i \cdot \mu_i)

    Where :math:`mu_i` is taken from the reference dataset.

    Parameters
    ----------
    adata : sc.AnnData
        Simulated annotated dataframe to be perturbed. Output from
        SingleCellDataset.simulate().
    samples : int, optional
        Number of perturbed cells to simulate, by default 200.
    pop_targets : list-like, optional
        Populations to simulate, by default None, and all populations present in
        `adata` will be simulated.
    gene_targets : list-like, optional
        Genes to perturb, by default None, and targets will be randomly chosen.
    percent_perturb : float, optional
        Percentage of genes to perturb. By default None, and if no argument is
        provided for `gene_targets`, will be set to 20%. If *both* arguments
        for `gene_targets` and `percent_perturb` are provided, the remainder of
        genes between `gene_targets` and the number of genes dictated by
        `percent_purturbed`, will be randomly selected.
    pop_sizes : list-like, optional
        Desired population sizes for each perturbed population. By default None,
        and the number of cells will be evenly distributed between populations.
    percentile : float, optional
        Percentile to use when calculationg dropout probabilities. Default is
        0.5, and the median will be used.
    new_populations : int, optional
        Number of unique populations to add to perturbed dataset. Default is 0.
    new_pop_pmarker : float, optional
        Probability of marker genes for added populations.
    
    Returns
    -------
    sc.AnnData
        Perturbed single-cell dataset.
    """
    adata.obs.Population = adata.obs.Population.astype(str)
    if new_pop_pmarker is not None:
        if not isinstance(new_pop_pmarker, (float, np.number)) and not (
            0 < new_pop_pmarker < 1
        ):
            raise ValueError("Expected marker probabilities between 0 and 1.")
    if new_pop_ids is not None and len(new_pop_ids) != len(new_pop_cells):
        raise ValueError(
            "Number of specified ids does not match number of " "new populations"
        )
    if perturbation_key is None:
        perturbation_key = "Perturbed"
    if new_pop_ids is None:
        new_pop_ids = [
            f"{perturbation_key}-added-{i + 1}" for i in range(len(new_pop_cells))
        ]
    if gene_targets is None:
        gene_targets = []
    else:
        try:
            gene_targets = list(gene_targets)
        except TypeError:
            raise ValueError("Expected list-like for `gene_targets`.")
    # get marker genes for reference dataset
    markers = population_markers(adata)
    if pop_targets is not None:
        pop_targets = utils.check_np_castable(pop_targets, "pop_targets")
        # ensure specified populations exist in provided adata
        if not set(pop_targets).issubset(adata.obs["Population"]):
            diff = set(pop_targets).difference(adata.obs["Population"])
            raise ValueError("Undefined populations: {}".format(diff))
        for each in pop_targets:
            gene_targets += list(markers[each])
    # check population sizes, if none, match with ratio in adata
    if pop_sizes is None:
        pop_counts = adata.obs["Population"].value_counts().values
        pop_sizes = (pop_counts / pop_counts.sum() * samples).astype(int)
        if samples % pop_sizes.sum() != 0:
            remainder = samples % pop_sizes.sum()
            iters = 0
            while remainder != 0:
                pop_sizes[iters % len(pop_sizes)] += 1
                remainder = samples % pop_sizes.sum()
                iters += 1
    else:
        pop_sizes = utils.check_np_castable(pop_sizes, "pop_sizes")
        if pop_sizes.sum() != samples:
            raise ValueError("Population sizes do not sum to number of samples.")
    if new_pop_cells is not None and len(new_pop_cells) > 0:
        new_pop_cells = utils.check_np_castable(new_pop_cells, "new_pop_cells")
        pop_sizes = np.hstack([pop_sizes, new_pop_cells])
        samples += sum(new_pop_cells)
    # determine which genes to perturb
    if gene_targets is not None:
        if not set(gene_targets).issubset(adata.var.index):
            raise ValueError(
                "Unrecognized gene targets: {}".format(
                    set(gene_targets).difference(adata.var.index)
                )
            )
        gene_targets = utils.check_np_castable(gene_targets, "gene_targets")
    n_genes = int(percent_perturb * adata.shape[1])
    # perturb remaining genes not specified by `gene_targets`.
    if len(gene_targets) < n_genes:
        n_genes -= len(gene_targets)
        ignore_genes = gene_targets
        for x in markers.values():
            ignore_genes = np.hstack((ignore_genes, x))
        p_genes = list(set(adata.var.index).difference(ignore_genes))
        targets = np.random.choice(p_genes, n_genes)
        gene_targets = np.hstack((targets, gene_targets))

    all_markers = set(np.hstack(list(markers.values())))
    disp_ = adata.var["Base.Dispersion"].values.reshape((-1, 1))
    var_ = adata.var.copy()
    var_["Perturbation.Shift"] = np.ones((adata.shape[1], 1))
    # shifting the preturbation by a large degree makes marker genes more
    # likely to dropout in perturbation dataset, pass dropout
    # setting alpha = 1, beta = 1 puts mean at (1 * 1) -- meaning on average
    # there won't be an increase or decrease in the read counts present in the
    # dataset, alpha=2, beta=2 shifts on average by 4, so multiply markers by
    # average to maintain signal
    a, b = 2, 2
    var_.loc[gene_targets, "Perturbation.Shift"] = stats.gamma(a=a, scale=b).rvs(
        size=gene_targets.size
    )
    # var_.loc[all_markers, 'Perturbation.Shift'] *= a * b

    populations = []
    sim_pops = adata.obs["Population"].unique()
    if len(new_pop_ids) > 0:
        sim_pops = np.hstack([sim_pops, new_pop_ids])

    n_pops = len(sim_pops)
    pop_columns = []
    pop_dropout = []
    possible_new_markers = set(adata.var.index) - all_markers - set(gene_targets)
    for i, each in enumerate(sim_pops):
        if each in new_pop_ids:
            print(f"Adding new population {each}...")
            name = str(each)
            n_markers = stats.binom(adata.shape[1], new_pop_pmarker).rvs(1)
            new_pop_markers = np.random.choice(list(possible_new_markers), n_markers)
            shifts = stats.gamma(3, 3).rvs(n_markers)
            var_[f"Pop.{name}.Mu"] = adata.var["Base.Mu"]
            # multiply MU values by (a * b) to maintain signal
            var_.loc[new_pop_markers, f"Pop.{name}.Mu"] *= shifts * (a * b)
            var_[f"Pop.{name}.Marker"] = False
            var_.loc[new_pop_markers, f"Pop.{name}.Marker"] = True
            possible_new_markers -= set(new_pop_markers)
            # asymmetrical population,
            pop_dropout.append(None)
        else:
            markers_i = markers[each]
            # log population identity in perturbed data
            # if marker gene is modified, populations is considered perturbed
            if len(set(markers_i).intersection(gene_targets)) != 0:
                name = "Perturbed-{}".format(each)
                var_[f"Pop.{name}.Marker"] = var_[f"Pop.{each}.Marker"]
            else:
                name = str(each)
            markers_idx = np.array([var_.index.get_loc(x) for x in markers_i])
            marker_dropout = var_.loc[markers_i, f"Pop.{each}.Dropout"].values
            pop_dropout.append(
                np.hstack([markers_idx.reshape(-1, 1), marker_dropout.reshape(-1, 1)])
            )

        pop_columns.append(f"Pop.{each}.Mu")
        # pop_dropout.append(f'Pop.{each}.Dropout')
        populations += [name] * pop_sizes[i]
    obs_ = pd.DataFrame(
        populations,
        columns=["Population"],
        index=["cell-{}".format(i + 1) for i in range(samples)],
    )
    # # calculate average expression values for each gene in each population
    # in perturbed dataset
    mus = (
        var_[pop_columns].values
        * np.ones((adata.shape[1], n_pops))
        * var_["Perturbation.Shift"].values.reshape(-1, 1)
    )
    X_, dropout, __ = simulate_counts(
        samples,
        mus,
        disp_,
        n_pops,
        pop_sizes,
        percentile=percentile,
        dropout=pop_dropout,
    )
    #   n_pops, pop_sizes)
    # perturb dropout
    for i, each in enumerate(sim_pops):
        var_[f"Pop.{each}.Dropout-Prtb"] = dropout[:, i]

    obs_["Treatment"] = perturbation_key
    adata = sc.AnnData(X=X_, obs=obs_, var=var_)
    adata.obs["Population"] = adata.obs["Population"].astype(str)
    return adata


def average_exp(scale_factor, n=1):
    r"""
    Simulate average expression parameters for simulated genes.

    Simulate average expression parameters for simulated genes. Average values
    are modelled using a scaled beta distribution.

    ..math:
        \mu_i \sim Beta(a, b) \cdot c

    Where :math:`a = 2`, :math:`b = 5`, and :math:`c` is a user provided scale
    factor.

    Parameters
    ----------
    scale_factor : float
        Scalar to multiply beta-distributed random variable to calculate
        expression averages.
    n : int, optional
        Number of genes to model, by default 1.

    Returns
    -------
    np.ndarray
        Average gene expression values.
    """
    return stats.beta(a=2, b=5).rvs(n) * scale_factor


def sigmoid(x):
    """
    Sigmoid function used to estimate probability of dropout.

    Parameters
    ----------
    x : float
        Value to pass to sigmoid function.

    Returns
    -------
    float
        Value of `x` passed through sigmoid function.
    """
    return 1 / (1 + np.e ** -x)


def dropout_probability(mu, median_avg, beta_0=-1.5):
    r"""
    Estimate the probability of dropout for a given gene.

    Estimate the probability of a dropout even using a sigmoid function, and
    the following models. 

    ..math::
        p_i = 1 - sigmoid(x) \\
        sigmoid(x) = \dfrac{1}{1 + e^{-x}} \\
        x_i = \beta_0 + \dfrac{mu_i}{median(\vec \mu)} \\
        \vec \mu = \{\mu_1, \mu_2, \ldots , \mu_{p - 1}, \mu_p}
    
    Parameters
    ----------
    mu : float, numpy.ndarray
        Either the average expression value for a single gene, or an array of
        average expression values for a set of genes.
    median_avg : float
        Median average expression value over all genes.
    beta_0 : float, optional
        Affine term to add to scalar multiple of `mu`. By default -1.5.
    
    Returns
    -------
    float
        Probability of gene(s) to experience a dropout event.
    """
    x = beta_0 + 1 / median_avg * mu
    return 1 - sigmoid(x)


def simulate_counts(
    n_samples,
    mus,
    dispersion,
    populations,
    pop_sizes,
    percentile=50,
    method="conditional",
    dropout=0.66,
):
    r"""
    Simulate counts across genes for a set number of samples.

    Parameters
    ----------
    n_samples : int
        Number of samples to simulate.
    mus : np.ndarray
        A gene by population matrix, where each element represents the average
        expression value in the given population.
    dispersion : numpy.ndarray
        A gene length vector with dispersion paramtersfor modelling counts
        using a negative binomial distribution for each gene.
    populations: int
        Number of populations to sample.
    pop_sizes : numpy.ndarray
        list of samples per population.
    percentile : float, optional
        Which percentile to use when calculating dropout probabilities. Default
        is 50, and the median will be used.
    method : str, optional
        How to perform dropout. Default is `conditional` and dropout rates for
        each gene will be conditioned on its average experession compared to the
        median average expression. The other option is `uniform` and all genes
        will share the same chance of dropping out. Default is conditional
    dropout : float, optional
        Value between 0 and 1 denoting the average chance of a dropout event for
        a given gene. Only used if `method=='uniform'`. Default is 0.66

    Returns
    -------
    (numpy.ndarray, numpy.ndarray)
        An :math:`n \times p` count matrix, where :math:`n` is number of
        samples, defined by `n_samples`, and :math:`p` is the number of genes,
        defined by the size of `mus`.

        A :math:`p` length vector of population labels for each row in returned
        count matrix.
    """
    # sample by gene expression matrix
    X_ = np.zeros((n_samples, mus.shape[0]))
    # vector labelling which population each sample belongs to
    labels_ = np.hstack([np.array([i + 1] * pop_sizes[i]) for i in range(populations)])
    # ensure proper shape of dispersion vector
    if len(dispersion.shape) == 1:
        dispersion = dispersion.reshape((-1, 1))
    elif len(dispersion.shape) != 2 and dispersion.shape[0] < dispersion.shape[1]:
        raise ValueError(
            "Expected column vector for dispersion. Received "
            "vector with shape {}.".format(dispersion.shape)
        )
    if len(mus.shape) == 1:
        mus = mus.reshape((-1, 1))
    # calculate theoretical expression averages across populations
    # |gene| x |populations| matrix
    means_ = mus * np.ones((dispersion.shape[0], populations)) * dispersion

    # calculate dropout probabilites for each gene in each population
    # a |gene| x |populations| size matrix
    if method == "conditional":
        if isinstance(dropout, np.ndarray):
            if (
                not dropout.shape[0] != means_.shape[0]
                and dropout.shape[1] != means_.shape[1]
            ):
                raise ValueError(
                    "When passing a numpy array for `dropout`, "
                    "assume a (|gene| x |population|) size array. "
                    f"expected {means_.shape}, got {dropout.shape}."
                )
            else:
                p_dropout = dropout
        else:
            percentile_ = np.percentile(means_, percentile)
            p_dropout = dropout_probability(means_, percentile_)
    elif method == "uniform":
        p_dropout = np.ones_like(means_) * dropout
    # marker gene dropouts were passed, conserve dropout probability between
    # controls + treated
    if isinstance(dropout, list):
        for i, prev_drop in enumerate(dropout):
            if prev_drop is not None:
                # dropout is gene x pop
                p_dropout[prev_drop[:, 0].astype(int), i] = prev_drop[:, 1]
    # simulate counts across populations
    for i in range(populations):
        if i == 0:
            start = 0
        else:
            start = pop_sizes[:i].sum()
        # simulate gene counts for each cell in population i
        for j in range(mus.shape[0]):
            dist = stats.nbinom(
                dispersion[j, 0], 1 - mus[j, i] / (mus[j, i] + dispersion[j, 0])
            )
            # calculate probability of reads not dropping out with bernoulli
            # 1 = keep count, 0 = dropeed, multiplying by dropout vector
            # moves dropped counts to 0.
            dropped = stats.bernoulli(p=1 - p_dropout[j, i]).rvs(pop_sizes[i])
            X_[start : start + pop_sizes[i], j] = dist.rvs(pop_sizes[i]) * dropped
    return X_, p_dropout, labels_


if __name__ == "__main__":
    ctrls = SingleCellDataset().simulate()
    prtbs = perturb(ctrls)
    sc.pp.pca(prtbs)
    sc.pl.pca_scatter(prtbs, color="Population")
    comb = ctrls.concatenate(prtbs)
    sc.pp.neighbors(comb)
    sc.tl.umap(comb)
    sc.pl.umap(comb, color="Treatment", palette=["black", "red"])
    sc.pl.umap(comb, color="Population")
    import icat

    model = icat.icat("Control", ncfs_kws={"sigma": 3, "reg": 3})
    clustered = model.cluster(comb, comb.obs["Treatment"])
    sc.tl.umap(clustered)
    sc.pl.umap(clustered, color="Population")



"""
Module to identify cell-types across treatments.

@author: Dakota Y. Hawkins
@contact: dyh0110@bu.edu

Example:

.. code-block::python
    from icat import simulate
    from icat import models
    import scanpy as sc
    import numpy as np
    data_model = simulate.SingleCellDataset(
        populations=2,
        genes=1000,
        dispersion=np.random.choice([1, 2, 3], 1000)
    )
    controls = data_model.simulate()
    controls.obs['treatment'] = 'control'
    perturbed = simulate.perturb(controls)
    perturbed.obs['treatment'] = 'perturbed'
    adata = controls.concatenate([perturbed])
    sc.pp.log1p(adata)
    model = models.icat(
        ctrl_value="control",
        ncfs_kws={'reg': 1, 'sigma': 3},
        neighbor_kws={'n_neighbors': 15}, 
        cluster_kws={'resolution': 0.75},
    )
    out = model.cluster(adata, adata.obs['treatment'])
    print(out.obs['sslouvain'].unique())
"""

import inspect
import warnings
import logging
import time

import numpy as np
import pandas as pd
from sklearn import preprocessing

from sklearn.neighbors import KNeighborsTransformer
import apricot
import scanpy as sc

import sslouvain
import ncfs

from . import utils


class icat:
    """
    Model to identify clusters across treatments.

    Parameters
    ----------
    ctrl_value : str
        Status label corresponding to control group in treatment observation
        data.
    clustering : str, optional
        Method to use for initial clustering of control cells, by default
        'louvain'. Options are 'louvain' or 'leiden'.
    treatment_col : str, optional
        Column name in oberservation data annotating treatment type between
        datasets, by default 'treatment'.
    reference : str, optional
        Which data partitions to use for when learning gene weights. Default is
        'all' and all partions by treatment type will be used (as defined by
        values in `treatment_col`). Acceptable values are 'all' and 'controls'.
    subsample : bool, optional
        Whether to perform NCFS training on a sample of the data. Default is
        False, and training will be performed on the entire dataset.
    train_size : float, optional.
        Proportion of data to train on if `subsample==True`. Values should be
        between 0 and 1. Default is 0.1, and training will be performed on 10%
        of the data.
    ncfs_kws : dict, optional
        Keyword arguments to pass to `ncfs.NCFS` model, by default None and
        default parameters are used. See `ncfs.NCFS` for more information.
    cluster_kws : dict, optional
        Keyword arguments to pass to clustering method, by default None and
        default parameters are used. See `scanpy.tl.louvain` or
        `scanpy.tl.leiden` for more information.
    cluster_col : str, optional
        Optional column in observation data denoting pre-identified clusters.
        By default None, and clustering will be performed following
        `clustering` and `cluster_kws`.
    neighbor_kws : dict, optional
        Keyword arguments for identifying neighbors. By default None, and
        default parameters are used. See `scanpy.pp.neighbors` for more
        information.
    sslouvain_kws : dict, optional
        Keyword arguments for `sslouvain`. By default None, and default
        parameters will be used. See `sslouvain.find_partition` for more
        information.
    pca_kws : dict, optional
        Keyword arguments for performing principle component analysis. By default
        None, and default parameters will be used. See `sslouvain.tl.pca` for
        more information.
    subsample_kws : dict, optional
        Keyword arguments for performing submodular optimization. By default
        None, and default parameters will be used. See `icat.select_cells` for
        more information.
    """

    def __init__(
        self,
        ctrl_value,
        clustering="louvain",
        reference="all",
        subsample=False,
        train_size=0.1,
        ncfs_kws=None,
        cluster_kws=None,
        cluster_col=None,
        neighbor_kws=None,
        sslouvain_kws=None,
        pca_kws=None,
        subsample_kws=None,
    ):
        self.ctrl_value = ctrl_value
        self.clustering = clustering
        self.reference = reference
        self.subsample = subsample
        self.train_size = train_size
        self.ncfs_kws = ncfs_kws
        self.cluster_kws = cluster_kws
        self.cluster_col = cluster_col
        self.neighbor_kws = neighbor_kws
        self.sslouvain_kws = sslouvain_kws
        self.pca_kws = pca_kws
        self.subsample_kws = subsample_kws
        # utils.set_log()

    @property
    def ctrl_value(self):
        return self._ctrl_value

    @ctrl_value.setter
    def ctrl_value(self, value):
        if not isinstance(value, str):
            raise ValueError("Expected string for `ctrl_value` parameter")
        self._ctrl_value = value

    @property
    def clustering(self):
        return self._clustering

    @clustering.setter
    def clustering(self, value):
        if not isinstance(value, str):
            raise ValueError(
                "Expected string for `clustering` parameter. "
                "Received {}".format(value)
            )
        value = value.lower()
        if value not in ["louvain", "leiden"]:
            raise ValueError("Unsupported method: {}".format(value))
        self._clustering = value

    @property
    def reference(self):
        """Reference data set. Acceptable values 'all' or 'controls'."""
        return self._reference

    @reference.setter
    def reference(self, value):
        if value not in ["all", "controls"]:
            raise ValueError("`reference` value should be either 'all' or 'controls'")
        self._reference = value

    @property
    def subsample(self):
        "Whether to perform NCFS training on a sample of the data."
        return self._subsample

    @subsample.setter
    def subsample(self, value):
        if not isinstance(value, bool):
            raise ValueError("Expected bool for `subsample`")
        self._subsample = value

    @property
    def train_size(self):
        "Proportion of data to train on. Default is 0.1"
        return self._train_size

    @train_size.setter
    def train_size(self, value):
        if not 0 < value < 1:
            raise ValueError("Expected value between 0 and 1 for `train_size`.")
        self._train_size = value

    @property
    def ncfs_kws(self):
        """
        Keyword arguments to pass to NCFS. See NCFS documentation for me information.
        """
        return self._ncfs_kws

    @ncfs_kws.setter
    def ncfs_kws(self, value):
        default_kws = utils.get_default_kwargs(ncfs.NCFS, ["self"])
        if value is not None:
            value = utils.check_kws(default_kws, value, "ncfs")
        else:
            value = default_kws
        self._ncfs_kws = value

    @property
    def neighbor_kws(self):
        """Keyword arguments for neighbor determination."""
        return self._neighbor_kws

    @neighbor_kws.setter
    def neighbor_kws(self, value):
        default_kws = utils.get_default_kwargs(sc.pp.neighbors, ["adata", "copy"])
        default_kws["metric"] = ncfs.distances.phi_s
        if value is not None:
            value = utils.check_kws(default_kws, value, "neighbor_kws")
        else:
            value = default_kws
        self._neighbor_kws = value

    @property
    def cluster_kws(self):
        """Keyword arguments to pass to clustering algorithm."""
        return self._cluster_kws

    @cluster_kws.setter
    def cluster_kws(self, value):
        default_kws = {
            "louvain": utils.get_default_kwargs(sc.tl.louvain, ["adata", "copy"])
        }
        if "leiden" in dir(sc.tl):
            default_kws["leiden"] = utils.get_default_kwargs(
                sc.tl.leiden, ["adata", "copy"]
            )
        else:
            if self.clustering == "leiden":
                raise ValueError(
                    "The version of Scanpy you are running does "
                    "not support Leiden clustering. If you wish "
                    "to use Leiden clustering, try upgrading your"
                    "version of Scanpy."
                )
        default_kws["resolution"] = 1.0
        if value is not None:
            value = utils.check_kws(
                default_kws[self.clustering], value, "cluster." + self.clustering
            )
        else:
            value = default_kws[self.clustering]
        self._cluster_kws = value

    @property
    def cluster_col(self):
        """Optional column identifying pre-computed clusters."""
        return self._cluster_col

    @cluster_col.setter
    def cluster_col(self, value):
        if not isinstance(value, (str, int)) and value is not None:
            raise ValueError(
                "Expected integer index or string name for " + "`cluster_col`"
            )
        self._cluster_col = value

    @property
    def sslouvain_kws(self):
        """Keyword arguments for sslouvain."""
        return self._sslouvain_kws

    @sslouvain_kws.setter
    def sslouvain_kws(self, value):
        default_kws = utils.get_default_kwargs(sslouvain.find_partition, "")
        if value is not None:
            value = utils.check_kws(default_kws, value, "sslouvain_kws")
        else:
            value = default_kws
        self._sslouvain_kws = value

    @property
    def pca_kws(self):
        """Keyword arguments for PCA."""
        return self._pca_kws

    @pca_kws.setter
    def pca_kws(self, value):
        default_kws = utils.get_default_kwargs(sc.pp.pca, ["data", "copy"])
        if value is not None:
            value = utils.check_kws(default_kws, value, "pca_kws")
        else:
            value = default_kws
        self._pca_kws = value

    @property
    def subsample_kws(self):
        """Keyword arguments for `select_cells()`"""
        return self._subsample_kws

    @subsample_kws.setter
    def subsample_kws(self, value):
        default_kws = utils.get_default_kwargs(
            self.select_cells, ["adata", "label_col"]
        )
        if value is not None:
            if "selector" in value:
                if not any(
                    [
                        value["selector"] == x
                        for x in [
                            apricot.MaxCoverageSelection,
                            apricot.MixtureSelection,
                            apricot.SumRedundancySelection,
                            apricot.FacilityLocationSelection,
                            apricot.FeatureBasedSelection,
                            apricot.GraphCutSelection,
                            apricot.SaturatedCoverageSelection,
                        ]
                    ]
                ):
                    raise ValueError(
                        "Expected `apricot` submodular "
                        "optimization method for `selector`."
                    )
            value = utils.check_kws(default_kws, value, "subsample_kws")
        else:
            value = default_kws
        self._subsample_kws = value

    # @profile
    def cluster(self, adata, treatment, verbose=False):
        """
        Cluster cells in control and experimental conditions.

        Parameters
        ----------
        adata : sc.AnnData
            Annotated dataframe of cells.
        perturbed : pd.Series, numpy.ndarray, list-like
            Treatment labels for each cell in `adata`.
        verbose : boolean, optional
            Whether to print progress through clustering step.

        Returns
        -------
        sc.AnnData
            Annotated dataframe of out cells in NCFS space.
        """
        self.log_ = False
        if self.log_:
            utils.set_log()
            utils.log_system_usage("Instantiation")

        self.verbose_ = verbose
        if not isinstance(adata, sc.AnnData):
            raise ValueError("Expected AnnData object for `controls`.")
        if isinstance(treatment, pd.Series):
            treatment = treatment.values
        else:
            treatment = utils.check_np_castable(treatment, "treatment")
        if not self.ctrl_value in treatment:
            raise ValueError(
                f"Control value {self.ctrl_value} not found " "in treatment array."
            )
        # if distance function is from ncfs.distances, expects feature weights
        # check to see if they were provided, otherwise set weights to 1
        metric = self.neighbor_kws["metric"]
        if not isinstance(metric, str) and metric.__module__ == "ncfs.distances":
            try:
                self.neighbor_kws["metric_kwds"]["w"]
            except KeyError:
                # check for dict to update previously specified values
                if isinstance(self.neighbor_kws["metric_kwds"], dict):
                    self.neighbor_kws["metric_kwds"].update(
                        {"w": np.ones(adata.shape[1])}
                    )
                else:
                    self.neighbor_kws["metric_kwds"] = {"w": np.ones(adata.shape[1])}
        if self.log_:
            utils.log_system_usage("Starting NCFS transformation.")
        self.__learn_weights(adata, treatment)

        # n_clusters = len(controls.obs[self.cluster_col].unique())
        adata.var["informative"] = adata.var["ncfs_weights"].apply(lambda x: x > 1)
        informative = adata.var["informative"].sum()
        if self.verbose_:
            print(f"Found {informative} informative features.")

        # return adata
        # if informative < n_clusters:
        #     warnings.warn("Number of informative genes less "
        #                   "than the number of identified control clusters: "
        #                   f"informative genes: {informative}, "
        #                   f"number of clusters: {n_clusters}. Consider "
        #                   "increasing `sigma` or decreasing `reg` for better "
        #                   "performance.")
        # scikit-learn 0.22, umap==0.4.4
        self.neighbor_kws["use_rep"] = "X_icat"
        if self.log_:
            utils.log_system_usage("Before NCFS neighbors.")
        # A = KNeighborsTransformer(mode='connectivity',
        #                           n_neighbors=self.neighbor_kws['n_neighbors'])\
        #                          .fit_transform(adata.obsm['X_icat'])

        sc.pp.neighbors(adata, **self.neighbor_kws)
        # sc.pp.neighbors(adata, n_neighbors=self.neighbor_kws['n_neighbors'])
        # sc.pp.neighbors(adata)
        if self.log_:
            utils.log_system_usage("After NCFS neighbors.")
        # grab connectivities of cells
        A = utils.get_neighbors(adata, "connectivities")
        g = utils.igraph_from_adjacency(A)
        # instantiate semi-supervised Louvain model
        try:
            resolution = self.cluster_kws["resolution"]
            if resolution is None:
                resolution = 1
        except KeyError:
            resolution = 1.0
        try:
            vertex = self.sslouvain_kws["partition_type"]
        except KeyError:
            vertex = sslouvain.RBConfigurationVertexPartition
        if not isinstance(vertex, utils.ig.VertexClustering):
            vertex = sslouvain.RBConfigurationVertexPartition
        adata.obs.loc[
            [x != self.ctrl_value for x in treatment], self.cluster_col_
        ] = None
        y_, mutables = utils.format_labels(adata.obs[self.cluster_col_])
        if self.verbose_:
            print(f"{sum(mutables)} cells set as mutable out of {len(y_)}")
        # logging.info("Runing sslouvain")
        part = sslouvain.find_partition(
            g,
            vertex,
            initial_membership=y_,
            mutable_nodes=mutables,
            resolution_parameter=resolution,
        )
        # store new cluster labels in cell metadata
        adata.obs["sslouvain"] = part.membership
        adata.obs["sslouvain"] = adata.obs["sslouvain"].astype("category")
        if self.verbose_:
            print("ICAT complete.")
        # utils.close_log()
        return adata

    # @profile
    def __learn_weights(self, adata, treatment):
        reference = [utils.subset_cells(adata, treatment, self.ctrl_value)]
        scaler = preprocessing.MinMaxScaler()
        if self.reference == "all":
            if self.verbose_:
                print("Using all datasets as reference sets.")
            reference += [
                utils.subset_cells(adata, treatment, x, False)
                for x in np.unique(treatment)
                if x != self.ctrl_value
            ]
            scaler.fit(adata.X)
        else:
            if self.verbose_:
                print("Using control samples as reference.")
            scaler.fit(reference[0].X)
        reference = self.__cluster_references(reference)
        model, weights = self.__ncfs_fit(reference, scaler)
        if self.verbose_:
            print("-" * 20 + " Finished NCFS Fitting " + "-" * 20)

        # set default label to None, supply reference cluster labels
        adata.obs[self.cluster_col_] = None
        # match selected cells in split reference datasets to combined
        if self.subsample:
            adata.obs["selected"] = None
        for ref in reference:
            adata.obs.loc[ref.obs.index, self.cluster_col_] = ref.obs[self.cluster_col_]
            if self.subsample:
                adata.obs.loc[ref.obs.index, "selected"] = ref.obs["selected"]
        del reference
        adata.obsm["X_icat"] = model.transform(scaler.transform(adata.X))

        # copy control clusters over to control dataset

        # save genes weights
        adata.var["ncfs_weights"] = model.coef_
        if self.reference == "all":
            for i, value in enumerate(np.unique(treatment)):
                adata.var["{}_weights".format(value)] = weights[i, :]
        # free up memory by deleting unecessary objects
        del model
        try:
            del adata.uns["neighbors"]
        except (AttributeError, KeyError):
            pass
        try:
            del adata.obsp["neighbors"]
        except (AttributeError, KeyError):
            pass
        if self.log_:
            utils.log_system_usage("NCFS transform complete.")

    # @profile
    def __cluster_references(self, reference):
        # logging.info('Clustering reference datasets.')
        # utils.log_system_usage()
        for i, ref in enumerate(reference):
            if self.verbose_:
                print("Clustering cells in reference {}.".format(i + 1))
                # logging.info("Clustering cells in reference {}".format(i + 1))
            # utils.log_system_usage()
            sc.pp.pca(ref, **self.pca_kws)
            sc.pp.neighbors(ref, **self.neighbor_kws)
            if self.cluster_col is not None:
                print(f"Using provided cell labels in {self.cluster_col}.")
                if self.cluster_col not in ref.obs.columns:
                    raise ValueError(
                        f"Provided cluster column {self.cluster_col} "
                        "not found in observation data."
                    )
                ref.obs[self.cluster_col + "-copy"] = ref.obs[self.cluster_col].copy()
                self.cluster_col_ = self.cluster_col + "-copy"
            else:
                if self.clustering == "louvain":
                    sc.tl.louvain(ref, **self.cluster_kws)
                    self.cluster_col_ = "louvain"
                elif self.clustering == "leiden":
                    sc.tl.leiden(ref, **self.cluster_kws)
                    self.cluster_col_ = "leiden"
        if self.log_:
            utils.log_system_usage("Clustered reference datasets.")
        return reference
        # logging.info('Cells clustered')
        # utils.log_system_usage()

    # @profile
    def select_cells(
        self,
        adata,
        label_col,
        method="submodular",
        selector=apricot.FacilityLocationSelection,
        use_rep="X",
        by_cluster=False,
        stratified=True,
    ):
        """
        Select cells to train NCFS weights on.

        Parameters
        ----------
        adata : sc.AnnData
            Annotated dataframe containing cell labels and expression values.
        label_col : str
            Column in `adata.obs` containing labels to use during NCFS.
        method : str, optional
            Method to select training cells. Possible values are 'submodular'
            and 'random'. By default 'submodular', and sobmodular optimization
            will be used to find the "best" cells to train on.
        selector : apricot.Function, optional
            Sub-optimal optimizer function from `apricot`. Default is
            `apricot.MaxCoverageSelection`.
        by_cluster : bool, optional
            Whether to perform submodular optimization on a per cluster basis.
            Otherwise perform on the whole dataset with labels provided. By
            default true.
        stratified : bool, optional
            Whether to select cells in a stratified manner. By default False,
            and for `by_cluster` submodular optimization, an equal number of
            cells from each cluster will be chosen. Otherwise, a number
            proportional to the label prevalence will be chosen.

        Returns
        -------
        (X_train, y_train) : (np.ndarray, np.ndarray)
            A tuple of the data matrix X and observation labels y to be used
            during NCFS training.
        """

        y = adata.obs[label_col].values
        if method == "submodular":
            train_X, train_y = utils.submodular_select(
                adata,
                y,
                by_cluster,
                stratified,
                self.train_size,
                self.neighbor_kws["metric"],
                selector,
                self.neighbor_kws["metric_kwds"],
                use_rep=use_rep,
                verbose=self.verbose_,
            )
        elif method == "random":
            train_X, train_y = utils.random_select(
                adata,
                y,
                self.train_size,
                stratified,
                use_rep=use_rep,
                verbose=self.verbose_,
            )
        elif method == "centroid":
            train_X, train_y = utils.centroid_collapse(
                adata, y, use_rep=use_rep, verbose=self.verbose_
            )

        elif method == "centroid_knn":
            train_X, train_y = utils.centroid_neighbors(
                adata,
                y,
                self.train_size,
                self.neighbor_kws["metric"],
                stratified,
                self.neighbor_kws["metric_kwds"],
                use_rep=use_rep,
                verbose=self.verbose_,
            )
        else:
            raise NotImplementedError(
                f"No support for method {method}. "
                "Supported methods are 'submodular' and "
                "random'."
            )
        if self.log_:
            utils.log_system_usage("After cell selection.")
        return train_X, train_y

    # @profile
    def __ncfs_fit(self, reference, scaler):
        # no previous clustering provided, cluster using louvain or leiden
        weights = np.zeros((len(reference), reference[0].shape[1]))
        for i, ref in enumerate(reference):
            X_train = ref.X
            y_train = ref.obs[self.cluster_col_].values
            if self.subsample:
                X_train, y_train = self.select_cells(
                    ref, self.cluster_col_, **self.subsample_kws
                )
            if self.verbose_:
                print(
                    f"Training NCFS on {X_train.shape[0]} samples out of "
                    + f"{ref.shape[0]}"
                )
            X_train = scaler.transform(X_train)
            # fit gene weights using control dataset
            if self.verbose_:
                start = time.time()
                print(
                    f"Starting NCFS gradient ascent for dataset {i + 1}/{len(reference)}"
                )

            # logging.info(f"Starting NCFS gradient ascent for dataset {i + 1}")
            # utils.log_system_usage()
            y_train, __ = utils.format_labels(y_train)
            model = ncfs.NCFS(**self.ncfs_kws)
            model.fit(X_train, y_train, sample_weights="balanced")
            weights[i, :] = model.coef_
            # logging.info(msg)
            if self.verbose_:
                msg = "Dataset {} NCFS complete: {} to convergence".format(
                    i + 1, utils.ftime(time.time() - start)
                )
                print(msg)
        model.coef_ = np.max(weights, axis=0)
        if self.log_:
            utils.log_system_usage("After NCFS fitting.")
        return model, weights



from setuptools import setup
import re

# taken from https://stackoverflow.com/questions/458550/standard-way-to-embed-version-into-python-package
VERSIONFILE = "icat/_version.py"
verstrline = open(VERSIONFILE, "rt").read()
VSRE = r"^__version__ = ['\"]([^'\"]*)['\"]"
mo = re.search(VSRE, verstrline, re.M)
if mo:
    verstr = mo.group(1)
else:
    raise RuntimeError("Unable to find version string in %s." % (VERSIONFILE,))

with open("README.md", "r") as f:
    long_description = f.read()

setup(
    name="icat-sc",
    version=verstr,
    description="Identify cell states across treatments in single-cell RNA sequencing experiments",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/BradhamLab/icat",
    author="Dakota Y. Hawkins",
    author_email="dyh0110@bu.edu",
    install_requires=[
        "ncfs",
        "sslouvain",
        "scikit-learn",
        "numpy",
        "scipy",
        "apricot",
        "pandas",
        "scanpy",
    ],
    license="BSD",
    packages=["icat"],
    zip_safe=False,
)
